/* $Id: drawxtl.h 1108 2011-01-23 18:10:46Z martin $ */

#define VERSION "4.5"
#define RAD 57.2957795		     /* Conversion from radians to degrees */
#define MAX_VERTS 20000		     /* Maximum number of generated vertices */
#define MXSYM 192                    /* Maximum number of symmetry operations */
#ifndef PI
#define PI 3.1415926
#endif
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define O_OPT 1
#define B_OPT 2
#define P_OPT 4
#define V_OPT 8
#define M_OPT 16
#define L_OPT 32
#define BUFSIZE 512
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define EPS 1e-7
#define SPIN 1
#define ZOOM 2
#define PAN 3
#define DRAG 4

typedef struct {
	float s;
	float v[3];
} QUAT;

typedef float XY[2];

typedef float XYZ[3];

#ifdef OPENGL
#if defined(__APPLE__) && defined(__MACH__) 
#  include <GLUT/glut.h>
#else
#  include <GL/glut.h>
#  include <GL/gl.h>
#endif
# if defined(FREEGLUT)
#  include <GL/freeglut_ext.h>
# endif 
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#ifdef WIN32
  #include <direct.h>
#endif


#ifndef max
#define max(a,b) ((a) >= (b) ? (a) : (b))
#define min(a,b) ((a) <= (b) ? (a) : (b))
#endif

/* prototype definitions */

void add_to_list (float *, int, int);
void add_vert (float *, int, int, int, int);
void add_vert_nc (float *);
void analyze_bonds (void);
void axeqb (double a[3][3], double b[3], double c[3]);
void axis_to_quaternion(float *axis, float theta, QUAT *quat);
void Blank_Strip (char *);
void build_box_contents (void);
void build_poly_list (int, int, int, int, float *);
void Calc_Rot (float *, float *);
int  check_atom_name(char *, char *);
void check_dynamic_storage(void);
int check_vert_alloc(int, int);
void Conv_Sym_Mat (void);
void convert_ellipsoid (void);
float convert_pos(char *);
void crystal_compute (int, int);
void delay (int);
float determinant (double x[3][3]);
float dist (int, int);
void do_drawing (void);
float dot0_3d(float,float,float,float,float,float,float,float,float);
void draw_cell (int);
void draw_GL_triple(void);
void draw_cursor (void);            /* display openGL cursor */
void Dump_View(void);
void expand_atom (int);
int  eigen (float *, float *, float *, float a[3][3]);
int end_flip(int);
float end_flip_real(float);
void Error_Box(char *);
int fillcube(unsigned char *, FILE *);
int find_atom(void);
int find_proj_atom(int, int);
void find_all_in_box (int);
void find_lattice_type( void);
void findsys(void);
void generate_arrows (void);
void generate_bonds (void);
void generate_cones (void);
void generate_ellipsoids (void);
void generate_gl_texts(void);
void generate_lsq_planes(void);
void generate_planes (void);
void generate_poly (void);
void generate_slab(void);
void generate_spheres (void);
void generate_texts (void);
void generate_triple(void);
void generate_aimsurf(void);
void calculate_voids (void);
void generate_voids(void);
void calc_simplevoids (void);
void generate_simplevoids(void);
void calculate_sas(void);
void calculate_msms(void);
void generate_msms(void);
void get_atom_id (void);
void get_input (int);
void get_label (char *, char *, char *, char *, char *, int);
int  get_next_token(char *, int, FILE *);
void getsym(char*, int, int);
void import_cif (char *, int, int, int *, int);
void import_fdat (char *, int, int, int);
void import_gsas (char *, int, int, int);
void import_pcr (char *, int, int, int);
void import_schakal (char *, int, int, int);
void import_shelx (char *, int, int, int);
void import_wien (char *, int, int, int);
void import_discus (char *, int, int, int);
void import_exc (char *, int, int, int);
void init_dynamic_storage(void);
void label_cell(void);
void Locate_Triple (void);
void make_bmat (int sys, float lat_con[6], double b_mat[3][3], float ginv[3][3],
                float rec_lat_con[6]);
float matinv(float a[3][3]);
void matmul(float a[3][3],float b[3][3],float c[3][3]);
int  Maximize_rho(int);
void Add_mapslice (int);
void modulate_parameters(float vert[3], double *occ, int sym_no, int atom_no);
int modulate_uij(float vert[3], int ellips_no, int atom_no, int sym_no);
void move_cursor(int, float);
int  not_in_slab(float,float,float);
float P_to_C (float);
void Output_Spheres (float *, int);
int  p_eigen (float *, float *, float x[3][3]);
void plot_vrml_poly (int);
float *polygon_normal_3d(int, float*);
float polygon_solid_angle_3d(int, float*, float*);
int position_cif(int, long, FILE *, const char *, char *);
char *dissect_symbol(const char *);
void print_map_info(int);
void print_sym (void);
void Progress_Window(int, char *, float);
void push_cylinder (float df[3], float at[3], float radius, char * color);
void quaternion_to_rotmatrix(QUAT *q, float *m);
void qmult(QUAT *q1, QUAT *q2, QUAT *dest);
void qnormalize(QUAT *q);
void read_dn6(char * infile, int);
void read_fcf(char * infile, int);
void read_m80(char * infile, int);
void read_m81(char * infile, int);
void read_flp(char * infile, int);
void read_grd(char * infile, int);
void read_inp (int);
void read_stf(char * infile, int);
void read_vasp(char * infile, int);
void read_w2k(char * infile, int);
void read_aim(char * infile, int);
void read_exc(char * infile, int);
void read_xsf(char * infile, int);
void sub_add_vert(float * vert, int no_cell);
void sub_add_vert_nc(float * vert, int no_cell);
void symop (char *);
void show_slab(void);
void skip_blocks(int i, FILE * in);
void tb_init(void);
float tb_project_to_sphere(float, float, float);
void Token_Strip (char *, int);
//int toupper (int);
void trackball(void);
void Transform_POV_Color(char *);
void Transform_VRML_Color(char *);
void trim_string(char string[], int len);
int Unique_Atom (void);
void Update_Cursor_List(float *, int);
void usage (void);
void vadd(float *src1, float *src2, float *dst);
float vdot(float *v1, float *v2);
void vcross(float *v1, float *v2, float *cross);
void vcopy(float *vsrc, float *vdst);
int vec_dif(int n1, float v1[3], int n2, float v2[3], int n3, float v3[3], float v[3]);
float vlength(float *v);
void vmult(float *src1, float *src2, float *dst);
float *vnew(float x, float y, float z);
void vnormalize(float *v);
void vprint(float *v);
void vscale(float *v, float f);
void vset(float *v, float x, float y, float z);
void vsub(float *src1, float *src2, float *dst);
void vzero(float *v);
void window_to_tb(float mx, float my, float *x, float *y);
QUAT XYZ_Rot_to_Q(float xrot, float yrot, float zrot);
void *zalloc(size_t size);
int getsys(int);
void findsys(void);
void getsym(char*, int, int);

typedef struct
{
    int info_valid;
    int map_type;
    char title[82];
    float lat_con[6];
    float rhomx;
    float rhomn;
    int map_int[3];
    float xlim[2];
    float ylim[2];
    float zlim[2];
    float x4lim[2];
    float x5lim[2];
    float x6lim[2];
    int res;
}  MAP_INFO;

struct arrow_struct
{
    int arrow_fn;        /* frame number for this arrow */
    char col_arrow[40];  /* the color of this one */
    float arrow_length;  /* the length of this arrow */
    float arrow_diam;    /*    and its diameter */
    float mag_xp[3];     /* position of arrow in real-space cell */
    float mag_xc[3];     /* components of arrow wrt XYZ */
};

struct cone_struct
{
    int cone_fn;         /* frame number for the cone commands */
    char cone_l1[5];     /* label for atom end of cone */
    int numlonepairs;    /* number of lone pairs to draw at atom */
    float cone_height;   /* height of cone */
    float cone_min;      /* cone tip radius */
    float cone_max;      /* cone base radius */
    char col_cone[40];   /* color and properties of cones for POV */
};

struct plane_struct
{
    int plane_fn;        /* frame number for plane commands */
    char plane_col[40];  /* plane colors for POV */
    char plane_l[5];     /* plane center label */
    float plane_size;    /* size of plane coordination */
};

struct poly_struct
{
    int poly_fn;            /* frame number for polyhedra commands */
    char poly_l[5];         /* polyhedral center label */
    char poly_t[5];         /* polyhedral vertex label */
    float poly_size;        /* size of polyhedron */
    float poly_min;         /* lower limit for shell-like polyhedron */
    char poly_col[40];      /* polyhedron colors for POV */
    float poly_rad_edge;    /* individual edge radius for polyhedron */
    char poly_col_edge[40]; /* color of individual poly edges for POV */
};

struct bplane_struct
{
    char bplane_t[10][5];   /* atom labels in plane list */
    int bplane_n[10];       /* atom numbers in plane list */
    int nbatoms;            /* number of items in list */
    float bplane_d1;        /* width of plane segment to draw */
    float bplane_d2;	    /* second dimension */
    char bplane_col[40];    /* color */
};

struct bond_struct
{
    int bond_fn;            /* frame number for bond commands */
    char bond_l1[5];        /* label for one end of bond */
    char bond_l2[5];        /* label for other end */
    float bond_size;        /* bond diameter */
    float bond_min;         /* minimum bond distance */
    float bond_max;         /* maximum bond distance */
    char col_bond[40];      /* color and properties of bonds for POV */
    int bond_style;	    /* solid bond or dash */
};

struct sphere_struct
{
    int sphere_fn;          /* frame number for sphere commands */
    char sphere_l[5];       /* sphere label */
    int sphere_n;           /* atom number - -1 for old behavior */
    float sphere_size;      /* sphere size */
    char sphere_col[40];    /* sphere colors */
};

struct label_struct
{
    int label_fn;             /* frame number */
    float label_x[3];         /* position of textual labels */
    char label_label[64];     /* label text */
};

struct map_struct
{
    int  FourierContourSolid;
    float FourierContourLevel;   /* Level for 3d, initial level for 2d */
    float FourierContourStep;    /* Height between contours for 2d */
    float FourierContourTop;     /* Upper level for 2d */
    char FourierContourColor[40];
    char FourierBackColor[40];
};

struct frame_struct            /* allocate at max_frame */
{
    float clip_lim[6];         /* limits of clip box in cell coords */
    float cryst_lim[6];        /* limits of view box in cell coords */
    float map_lim[9];
    int map_lim_set;
    int slice;
    float mapslice[3];
    float mapnorm[3];
    float planeeq[4];
    char molecule_atom_name[5];
    int molecule_atom_no;
    float molecule_distance;
};

struct atom_struct             /* allocate at max_atom */
{
    float atom_xyz[3];         /* atomic coordinates */
    char atom_l[5];            /* atom labels */
    int atom_n;                /* input atom number */
    int atom_fn;               /* frame number for atom */
    int sv_atom_n;             /* saved input atom number */
    int   atom_ismod;          /* flag for modulated position */
    int   occ_ismod;           /* flag for modulated occupancy */
    float occupancy;           /* average occupancy */
    float min_occ;             /* minimum occupancy required for drawing the atom */
    float saved_xyz[3];        /* storage to save the atom position */
    int TF_status;             /* Temperature factor status (-1 no Uij, 0 sphere, 1 ellipsoid) */
    float radius;              /* vdW Radius for solvent calculations etc. */
};

struct ellips_struct           /* allocate at max_el */
{
  char ellips_col[40];         /* ellipsoid colors for POV */
  float ellips_EV[3][3];       /* Eigenvectors */
  float ellips_RMS[3];         /* RMS amplitudes */
  char ellips_l[5];            /* label for anisotropic atom */
  int ellips_n;                /* atom number */
  int save_el_number;          /* save original atom number */
  int ell_type;                /* form of ellipsoid input */
  float ellips[6];             /* thermal ellipsoid coefficients */
  int ellips_ismod;            /* flag for modulated Uij */
  char el_color_tmp[256];      /* temporary color array */
};

struct mod_gbl_struct          /* allocate at max_modulation */
{
  int   vector_mult[3];        /* coefficients ijk for modvector = i*cell_vec_1 + j*cell_vec_2 .. */
  float modvector[3];          /* atom Fourier modulation vectors */
};

struct mod_x_struct           /* allocate at max_atom * max_modulation */
{
  float atom_mod_sawtooth[5]; /* amplitudes and interval for sawtooth modulation */
  int   ellips_modpar_atom;   /* atom number for this item */
  float atom_occpar[2];       /* cosine and sine multiplier of occ fourier series */
  int   atom_occpar_atom;     /* atom number for atom_occpar */
  int   atom_occpar_id;       /* axis number (0-2) for atom_occpar */
  float atom_occ_crenel[2];   /* crenel lower and upper limit for occupancy */
};

struct mod_3x_struct          /* allocate at 3 * max_atom * max_modulation */
{
  float atom_modpar[2];       /* cosine and sine multiplier for an item */
  int   atom_modpar_atom;     /* atom number for atom_modpar */
  int   atom_modpar_axis;     /* axis number (0-2) for atom_modpar */
  int   atom_modpar_id;       /* entry in vector_mult for this term (1 - no_mod_vectors) */
};

struct mod_3t_struct          /* allocate at 3 * max_el * max_modulation */
{
  float ellips_modpar[2];     /* wave vector id, cosine and sine multip */
  int   ellips_modpar_id;     /* entry for this term (1 - no_mod_vectors) */
  int   ellips_modpar_term;   /* identity of this term (1-6); */
};

struct atprop_struct           /* atomic properties */
{
  int atprop_fn;             /* frame number */
  char atprop_l[5];          /* label */
  int atprop_n;              /* atom number - -1 for old behavior */
  float radius;              /* van der waals size */
};


typedef struct
{
/* global variables for lattice constants and symmetry operators */

    char spg[3][4];	          /* space group code */
    int ss[MXSYM][3][3];          /* rotation operators for symmetry */
    float rss[MXSYM][3][3];       /* unitary rotation operators for symmetry */
    float rssC[MXSYM][3][3];      /* crystallographic rotation operators for symmetry */
    float lat_pos[4][3];          /* coordinates of lattice points */
    float ts[MXSYM][3];           /* translation operators */
    int sym_op_no[MXSYM];         /* saved symmetry operation number */
    int sym_op_signed[MXSYM];     /* saved symmetry operation number with sign */
    float cell_xyz[MXSYM][3];     /* storage for generated equiv. positions */
    int sys;                      /* crystal system indicator */
    int nbr;                      /* Bravais lattice indicator */
    int acentric;                 /* true in non-centric space group */
    int ng;	                  /* number of rotation operations to specify group */
    int nlat;                     /* number of lattice positions */
    int origin1_flag;             /* flag for origin1 space group */
    float origin_offset[3];
    float label_scale;
    char col_cell[40];            /* color and properties of framebars */

/* global variables for file names and file variables */

    char savedpath[100];        /* path at start of program */
    char File_Root[100];        /* Input File Name with extension stripped */
    FILE *fpin, *flout;
    FILE *fpoutv;
    FILE *fpoutp;
    FILE *fpouta;
    char *Str_File_Inp;         /* Input File Name */
    char *Str_File;
    char Pov_File[100];
    char *output;
    char MSMS_Path[1024];
    
/* global variables for atom lists */

    int verts_alloc;            /* number of vertices allocated */
    struct atom_struct *atoms;  /* pointer where atom data is stored */
    int atom_alloc;             /* number of atoms allocated */
    int *atom_no;               /* master list atom and symmetry number */
    int *atom_so;               /* atom symmetry number including inversion*/
    int *orig_atom_no;          /* saved atom number */
    float xyzoff[3];            /* amount to be added to each atom position */

/* atom list variables for modulated structures */

  struct mod_gbl_struct *modulate_gbl; /* pointer to gloabal modulation parameters */
  struct mod_x_struct *modulate_x;     /* pointer to X modulation parameters */
  struct mod_3x_struct *modulate_3x;   /* pointer to 3X modulation parameters */
  struct mod_3t_struct *modulate_3t;   /* pointer to TF modulation */
  int mod_gbl_alloc;                   /* number of global modulations allocated */
  int mod_x_alloc;                     /* number of X modulations allocated */
  int mod_3x_alloc;                    /* number of 3X modulations allocated */
  int mod_3t_alloc;                    /* number of 3T modulations allocated */
  int modulated;                       /* flag for modulated or conventional/average structure */
  int no_mod_vectors;                  /* number of atom_site_fourier_wave_vectors */
  int no_site_displace;                /* number of atom_site_displace_fourier terms */
  int no_site_U_terms;                 /* number of atom_site_U_fourier_terms */
  int no_site_occ;                     /* number of atom_site_occ_fourier terms */
  int no_cell_vec;                     /* number of cell_wave_vectors */
  float cell_vec[3][3];                /* cell_wave_vectors */
  int no_subsys;                       /* number of subsystems */
  float subsys_ref_volume;             /* reference reciprocal volume */
  float subsys_vol[10];                /* reciprocal volume for each of the subsystems */
  float subsys_fact[10][3][3];         /* place to save relationship between main and subsystem cells */
  float phaseshift[3];                 /* modulation phase offset t for each modulation */
  float * vert_occ;                    /* occupancy at individual sites (for radius scaling) */
  float ts_m[24][3];                   /* translational component of superspace symmetry */
  int ss_m[24][3][3];                  /* rotational component of superspace symmetry */

/* cone parameters */

   int ncone;                      /* number of different types of lonepair cones */
   int ncone_alloc;                /* number of lonepairs currently allocated */
   struct cone_struct *cones;      /* place to allocate cone space */

/* magnetic arrow parameters */

  int nmag;                       /* number of magnetic */
  int nmag_alloc;                 /* number of moments currently allocated */
  float mag_matrix[3][3];         /* transformation between magnetic and nuclear cell */
  struct arrow_struct *arrows;    /* place to allocate arrow space */

/* polyhedra parameters */

  int npoly;                      /* number of different types of polyhedra */
  int npoly_alloc;	          /* number of polyhedra that have been allocated */
  float polylimit;                /* should be close to zero to allow only regular polygonal faces */
  struct poly_struct *polyhedra;  /* place to allocate polyhedra space */
  char col_edge[40];              /* default color and properties of polyhedron edges */
  float rad_edge;                 /* default radius for edge */

/* plane parameters */

  int nplane;                      /* number of different types of planes */
  int nplane_alloc;                /* number of planes allocated */
  struct plane_struct *planes;     /* place to allocate space for planes */

/* bond parameters */

  int nbond;                       /* number of different types of bonds */
  int nbond_alloc;                 /* number of bond types allocated */
  struct bond_struct *bonds;       /* pointer to bond storage */
  float Bond_Mult;                 /* Multiplier for Bond Size */
  float BndMult;                   /* Bond multiplier initialization */
  float mol_d;                     /* intramolecular distance for molecular and polyhedral completion */

/* text label parameters */

  int nlabel;                        /* number of textual labels */
  int nlabel_alloc;                  /* amount of space allocated */
  struct label_struct *labels;       /* pointer to label storage */
  int labels_moved;
  int autolabel;                     /* flag for autolabel of SHELX atoms */
  int triple[4];

/* least squares plane parameters */

  int nbplane;                       /* number of least squares planes */
  int nbplane_alloc;                 /* number of least-squares planes allocated */
  struct bplane_struct *bplanes;     /* pointer to allocated storage */

/* sphere parameters */

  int nsphere;                       /* number of different types of spheres */
  int nsphere_alloc;                 /* number allocated */
  struct sphere_struct *spheres;     /* pointer to sphere allocation */
  float Sphere_Mult;                 /* Multiplier for Sphere Size */
  float SpMult;                      /* initialization for Sphere_Mult */


/* background parameters */

  char col_bg[40];             /* Background color */
  float glback[3];             /* Background RGB values for openGL */

/* ellipsoid parameters */

  int auto_ellipse;
  struct ellips_struct *ellips;  /* pointer to where ellipsoid data are stored */
  int ellips_alloc;              /* number of ellipsoids allocated */
  float Phong_Value;             /* reflectivity value for Phong highlighting */
  float Phong_Size;              /* Phong highlighting size */
  char Cutout_color[40];         /* color of cutout section */
  int n_ellips;			 /* number of ellipsoids */
  int do_ellipsoids;             /* non-zero if ellipsoids to be drawn */
  int El_Cutout;                 /* True if ellipsoids to have an octant removed */
  char Ellipaxis_color[40];      /* color of principal axes */
  float Ellipsoid_Prob;          /* Fraction of electron dens enclosed by ellipsoid */
  float Ellipsoid_Scale;         /* Scaling needed to get probability */
  float Ellipaxis_width;         /* linewidth of principal axes */

/* lattice constant variables */
  
  float lat_con[6];          /* lattice constants */
  float rec_lat_con[6];      /* reciprocal lattice constants */
  double b_mat[3][3];        /* metric tensor */
  float ginv[3][3];          /* direct space metric tensor */
  float b_inv_mat[3][3];      /* inverse metric tensor */

/* lattice parameters of cutout box */

  float slab_con[6];         /* lattice constants */
  float slab_off[3];         /* offset */
  float slab_rot[3];         /* rotation angles */

/* Fourier Map parameters */

  int numOfFourierContours;
  int num_Fourier_alloc;          /* number allocated */
  struct map_struct *fourier;     /* pointer to dynamic Fourier storage */
  int Fourier2d;                 /* true if 2d Fourier */
  int Slice;
/* Bader Surface parameters */

#define MAX_SURF 5
  int nsurf;
  char surfatom[MAX_SURF][5];
  char surfcolor[MAX_SURF][40];
  int surfnum[MAX_SURF];
  int surftype[MAX_SURF];
  int ntet[MAX_SURF];
  int nphi[MAX_SURF];
  float *surfx[MAX_SURF];
  float *surfy[MAX_SURF];
  float *surfz[MAX_SURF];

/* cavity or solvent accessible surface parameters */
  char ***voidmap;
  int voidgrid[3];
  float probesize;
  char voidcolor[40];
  signed char voidflag;

/* atomic property parameters */

  int natprop;                /* number of different types of property definitions */
  int natprop_alloc;          /* number allocated */
  struct atprop_struct *atprops;      /* pointer to allocation */
              
/* variables used in drawing */

    int frame_no;
    int max_frame;              /* maximum number of frames */
    int frame_alloc;            /* number of frames allocated */
    struct frame_struct *frames;/* pointer to frame data */
    float X_Boxlim;             /* X box limit read from command line */
    float Y_Boxlim;             /* Y box limit read from command line */
    float Z_Boxlim;             /* Z box limit read from command line */
    float X_Origin;             /* X origin read from command line */
    float Y_Origin;             /* Y origin read from command line */
    float Z_Origin;             /* Z origin read from command line */
    float lookat_v1[3], lookat_v2[3];/* "lookat" vectors " */
    int noshadow;               /* True if no shadows for POV objects */
    float ambient;              /* POV lighting options */
    float diffuse;
    float specular;
    float roughness;

/* stereo parameters */

    int Stereo;                 /* TRUE if stereo pair to be generaged */
    int cross_eyed;             /* TRUE if "cross eyed" view */
    float stereo_base;          /* camera separation */

    int automation;
} CRYSTAL;

void Convert_Cart_Cryst(float bmat_inv[3][3], float inp[3], float oup[3], float origin[3]);

void Convert_Cryst_Cart(double b_mat[3][3], float inp[3], float oup[3], float origin[3]);
